# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

intents = {
    "Facility Information": [
        "Types of sports facilities available",
        "Operating hours and location",
        "Equipment rental and usage",
        "Accessibility and special accommodations",
    ],
    "Membership and Registration": [
        "Membership options and benefits",
        "Registration process for activities",
        "Membership renewal and cancellation",
        "Family and group membership plans",
    ],
    "Programs and Activities": [
        "Fitness class schedules",
        "Sports leagues and tournaments",
        "Recreational activities for all ages",
        "Seasonal programs and camps",
    ],
    "Personal Training and Coaching": [
        "Personal trainer services",
        "Sport-specific coaching",
        "Fitness assessment and goal setting",
        "Group training sessions",
    ],
    "Event Hosting and Parties": [
        "Facility rental for events",
        "Birthday party packages",
        "Corporate events and team building",
        "Sporting event organization",
    ],
    "Health and Wellness": [
        "Nutrition and diet consultation",
        "Wellness workshops and seminars",
        "Health screenings and assessments",
        "Mind-body fitness programs",
    ],
    "Youth Sports and Activities": [
        "Youth sports programs",
        "After-school activities",
        "Summer camps for kids",
        "Youth development initiatives",
    ],
    "Safety and Emergency Procedures": [
        "Safety guidelines and policies",
        "Emergency response procedures",
        "First aid and CPR training",
        "Water safety and lifeguard services",
    ],
    "Special Offers and Promotions": [
        "Discounts and special rates",
        "Referral and loyalty programs",
        "Seasonal promotions",
        "Gift cards and vouchers",
    ],
    "Community Engagement": [
        "Community outreach events",
        "Volunteer opportunities",
        "Local sports and recreation partnerships",
        "Charity and fundraising events",
    ],
    "Feedback and Improvement": [
        "Member feedback and surveys",
        "Suggestions for new programs or services",
        "Complaint resolution process",
        "Testimonials and success stories",
    ],
    "Outdoor and Adventure Sports": [
        "Outdoor recreation areas",
        "Adventure sports programs",
        "Guided outdoor excursions",
        "Equipment and safety for outdoor activities",
    ],
    "Aquatic Programs": [
        "Swimming pool schedule",
        "Aquatic fitness classes",
        "Swim lessons for all ages",
        "Lifeguard training and certification",
    ],
}
