# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

models = {
    # fmt: off
    "bart": "facebook/bart-large-cnn",
    "distillbart": "sshleifer/distilbart-cnn-12-6",
    "flan": "philschmid/flan-t5-base-samsum",
    "pegasus": "google/pegasus-cnn_dailymail",
    "led": "pszemraj/led-large-book-summary",
    "food": "Dizex/FoodBaseBERT-NER",
    "disease": "pruas/BENT-PubMedBERT-NER-Disease",
    # fmt: on
}


# facebook/bart-large-cnn
# sshleifer/distilbart-cnn-12-6
# philschmid/flan-t5-base-samsum
# google/pegasus-xsum
# pszemraj/led-large-book-summary
# knkarthick/MEETING-SUMMARY-BART-LARGE-XSUM-SAMSUM-DIALOGSUM-AMI
# ainize/kobart-news
# human-centered-summarization/financial-summarization-pegasus
# Callidior/bert2bert-base-arxiv-titlegen
# pszemraj/long-t5-tglobal-base-16384-book-summary
# Quake24/easyTermsSummerizer
# facebook/bart-large-cnn


# google/bigbird-pegasus-large-arxiv
# google/bigbird-pegasus-large-bigpatent
# google/bigbird-pegasus-large-pubmed
# google/flan-t5-base
# google/flan-t5-large
# google/flan-t5-xl
# google/flan-t5-xxl
# google/t5-v1_1-base
# google/t5-v1_1-large
# google/t5-v1_1-xl
# google/t5-v1_1-xxl
# google/umt5-base
# google/umt5-xl
# google/umt5-xxl
# google/pegasus-x-base
# google/pegasus-large
# google/pegasus-x-large
# google/pegasus-multi_news
