# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# MoritzLaurer/mDeBERTa-v3-base-xnli-multilingual-nli-2mil7
# google/t5_xxl_true_nli_mixture
# facebook/bart-large-mnli
# microsoft/deberta-v2-xlarge-mnli
# khalidalt/DeBERTa-v3-large-mnli
# typeform/distilbert-base-uncased-mnli
# roberta-large-mnli
# microsoft/deberta-v2-xxlarge-mnli
# sileod/deberta-v3-large-tasksource-nli
# cross-encoder/nli-deberta-v3-small
# cross-encoder/nli-deberta-v3-base
# cross-encoder/nli-deberta-v3-large
# cross-encoder/nli-roberta-base
# cross-encoder/nli-deberta-v3-large
# cross-encoder/nli-distilroberta-base
