# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .artisan_marketplace import intents as artisan_marketplace_intent
from .automotive import intents as automotive_intent
from .banking import intents as banking_intent
from .business_directory import intents as business_directory_intent
from .childcare import intents as childcare_intent
from .cultural import intents as cultural_intent
from .customer_support import intents as customer_support_intent
from .diet import intents as diet_intent
from .ecommerce import intents as ecommerce_intent
from .education import intents as education_intent
from .employee_helpdesk import intents as employee_helpdesk_intent
from .environment import intents as environment_intent
from .esoteric import intents as esoteric_intent
from .events import intents as events_intent
from .fitness import intents as fitness_intent
from .food_ordering import intents as food_ordering_intent
from .freelancing import intents as freelancing_intent
from .gardening import intents as gardening_intent
from .government import intents as government_intent
from .healthcare import intents as healthcare_intent
from .hobby import intents as hobby_intent
from .library import intents as library_intent
from .news import intents as news_intent
from .non_profit import intents as non_profit_intent
from .performing_arts import intents as performing_arts_intent
from .personal_finance import intents as personal_finance_intent
from .pet_care import intents as pet_care_intent
from .public_safety import intents as public_safety_intent
from .real_estate import intents as real_estate_intent
from .religion import intents as religion_intent
from .senior_care import intents as senior_care_intent
from .social_services import intents as social_services_intent
from .sports_centers import intents as sports_centers_intent
from .tourist_guide import intents as tourist_guide_intent
from .translation import intents as translation_intent
from .travel import intents as travel_intent
from .wedding import intents as wedding_intent
from .wellness import intents as wellness_intent
