# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

intents = {
    "Course Information": [
        "Course offerings",
        "Course schedule and timings",
        "Credit and grading system",
        "Prerequisites for courses",
    ],
    "Enrollment and Registration": [
        "Enrollment process",
        "Class registration",
        "Waitlist inquiries",
        "Dropping a course",
    ],
    "Academic Guidance": [
        "Major and minor selection",
        "Academic advising",
        "Study abroad options",
        "Internship opportunities",
    ],
    "Examinations and Assessments": [
        "Exam schedule",
        "Assignment submissions",
        "Grading queries",
        "Exam preparation resources",
    ],
    "Tuition and Financial Aid": [
        "Tuition fees",
        "Scholarships and grants",
        "Student loan information",
        "Payment plans",
    ],
    "Student Services": [
        "Counseling services",
        "Career center",
        "Health services",
        "Student clubs and organizations",
    ],
    "Technical Support": [
        "Learning management system help",
        "Technical issues with online resources",
        "Accessing digital libraries",
        "Software and hardware resources",
    ],
    "Faculty and Staff Information": [
        "Contacting faculty",
        "Office hours",
        "Research and publications",
        "Staff directory",
    ],
    "Campus Facilities": [
        "Campus amenities",
        "Library services",
        "Sports facilities",
        "Accommodation and housing",
    ],
    "Policies and Regulations": [
        "Academic integrity policy",
        "Attendance rules",
        "Student conduct code",
        "Complaint and grievance procedures",
    ],
    "Alumni Services": [
        "Alumni events and reunions",
        "Networking opportunities",
        "Continuing education",
        "Alumni benefits and services",
    ],
    "Remote and Online Learning": [
        "E-learning resources",
        "Remote class participation",
        "Online course access",
        "Blended learning support",
    ],
}
