# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

intents = {
    "Childcare Services": [
        "Finding local childcare",
        "Childcare service options",
        "Childcare costs and subsidies",
        "After-school care programs",
    ],
    "Parenting Advice": [
        "Newborn care tips",
        "Handling toddler behavior",
        "Teen parenting guidance",
        "Balancing work and parenting",
    ],
    "Health and Nutrition": [
        "Child nutrition and diet",
        "Vaccination schedules",
        "Dealing with common illnesses",
        "Mental health resources for children",
    ],
    "Education and Learning": [
        "Early childhood education",
        "School enrollment procedures",
        "Home-schooling resources",
        "Learning disability support",
    ],
    "Activities and Recreation": [
        "Family-friendly activities",
        "Children's clubs and classes",
        "Outdoor and educational games",
        "Summer camps and workshops",
    ],
    "Safety and Emergency Procedures": [
        "Child safety measures",
        "Emergency contacts for children",
        "First aid for kids",
        "Preparing children for emergencies",
    ],
    "Product and Gear Recommendations": [
        "Baby gear and product advice",
        "Reviews on children's products",
        "Age-appropriate toys and books",
        "Car seat and stroller recommendations",
    ],
    "Behavioral and Developmental Support": [
        "Child developmental milestones",
        "Behavioral therapy options",
        "Support for special needs",
        "Coping with developmental disorders",
    ],
    "Family Support Services": [
        "Accessing family counseling",
        "Support groups for parents",
        "Financial aid for families",
        "Legal advice for family issues",
    ],
    "Parental Leave and Work-Life Balance": [
        "Parental leave policies",
        "Returning to work after childbirth",
        "Work-life balance tips",
        "Flexible work arrangements",
    ],
    "Community and Networking": [
        "Parenting groups and forums",
        "Local family events",
        "Connecting with other parents",
        "Volunteering with children",
    ],
}
