# 🧠 Geniusrise
# Copyright (C) 2023  geniusrise.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

intents = {
    "Plant Care and Maintenance": [
        "Watering and feeding schedules",
        "Pruning and trimming techniques",
        "Pest and disease management",
        "Seasonal plant care tips",
    ],
    "Garden Design and Landscaping": [
        "Landscape design ideas",
        "Choosing plants for garden themes",
        "Hardscaping and garden structures",
        "Sustainable and eco-friendly gardening",
    ],
    "Indoor and Urban Gardening": [
        "Houseplant care and selection",
        "Setting up an indoor garden",
        "Balcony and rooftop gardening",
        "Container and small space gardening",
    ],
    "Soil and Fertilization": [
        "Soil testing and improvement",
        "Organic and inorganic fertilizers",
        "Composting methods",
        "Mulching techniques",
    ],
    "Irrigation and Watering Systems": [
        "Irrigation system installation",
        "Water conservation in gardening",
        "Drip irrigation and soaker hoses",
        "Automated watering solutions",
    ],
    "Vegetable and Herb Gardening": [
        "Growing vegetables at home",
        "Herb garden setup and care",
        "Edible plants for gardens",
        "Seasonal vegetable gardening",
    ],
    "Flower Gardening": [
        "Annual and perennial flowers",
        "Bulb planting and care",
        "Flower garden design",
        "Attracting pollinators",
    ],
    "Trees and Shrubs Care": [
        "Tree planting and care",
        "Shrub selection and pruning",
        "Fruit tree maintenance",
        "Hedge and border plants",
    ],
    "Garden Pests and Diseases": [
        "Identifying and controlling pests",
        "Dealing with plant diseases",
        "Organic pest control methods",
        "Beneficial insects and natural predators",
    ],
    "Workshops and Educational Programs": [
        "Gardening workshops and classes",
        "Horticulture training programs",
        "Gardening for beginners",
        "Specialized gardening techniques",
    ],
    "Tools and Equipment": [
        "Gardening tool recommendations",
        "Equipment maintenance and repair",
        "New gardening technologies",
        "Renting vs. buying equipment",
    ],
    "Sustainability and Eco-Gardening": [
        "Creating a wildlife-friendly garden",
        "Native plants and biodiversity",
        "Rain gardens and stormwater management",
        "Organic gardening practices",
    ],
    "Consultation and Professional Services": [
        "Landscape consultation",
        "Garden planning and design services",
        "Professional gardening services",
        "Horticultural therapy",
    ],
}
